#
# test qsort binding and callback implementation
#
lappend auto_path .

package require Ffidl
package require Ffidlrt

source qsort.tcl

#
# test for callbacks
#
if {[llength [info commands ::ffidl::callback]] == 0} return

#
# I use these, in error, so often that I should just define them
#
proc ::ffidl::format {type} { ::ffidl::info format $type }
proc ::ffidl::sizeof {type} { ::ffidl::info sizeof $type }

#
# compare ints given pointers
#
proc cmp.int {p1 p2} {
    binary scan [::ffidl::peek $p1 [::ffidl::sizeof int]] [::ffidl::format int] v1
    binary scan [::ffidl::peek $p2 [::ffidl::sizeof int]] [::ffidl::format int] v2
    #puts "cmp.int $p1 $p2 -> $v1 $v2 -> [expr {$v1-$v2}]"
    expr {$v1-$v2}
}

#
# define callback
#
::ffidl::callback cmp.int {pointer pointer} int

#
# display a list of ints
#
proc print.ints {ints} {
    foreach {i1 i2 i3 i4 i5 i6 i7 i8} $ints {
	puts "$i1 $i2 $i3 $i4 $i5 $i6 $i7 $i8"
    }
}

#
# construct a binary array of random ints
#
set n 128
for {set i 0} {$i < $n} {incr i} {
    lappend ints [expr {int(1000*rand())}]
}
#puts before
#print.ints $ints
set lsorted [lsort -integer $ints]

set ints [binary format [::ffidl::format int]* $ints]

#
# qsort them
#
qsort ints $n [::ffidl::sizeof int] cmp.int

#
# convert back to Tcl
#
binary scan $ints [::ffidl::format int]* ints

#
# write them out
#
#puts after
#print.ints $ints

if {[string compare $lsorted $ints] != 0} {
    puts "lsorted list:"
    print.ints $lsorted
    puts "qsorted list:"
    print.ints $ints
}
