#
# photo.tcl - access photo images via ffidl
#
package require Ffidl 0.1
package require Ffidlrt 0.1

# define a type alias for Tk_PhotoHandle
ffidl-typedef Tk_PhotoHandle pointer

# define a structure for Tk_PhotoImageBlock
ffidl-typedef Tk_PhotoImageBlock pointer int int int int int int int int

# bind to tk
ffidl-proc ffidl-find-photo {pointer pointer-utf8} Tk_PhotoHandle \
    [ffidl-symbol [ffidl-find-lib tk] Tk_FindPhoto]
ffidl-proc ffidl-photo-put-block {Tk_PhotoHandle pointer-byte int int int int} void \
    [ffidl-symbol [ffidl-find-lib tk] Tk_PhotoPutBlock]
ffidl-proc ffidl-photo-put-zoomed-block {Tk_PhotoHandle pointer-byte int int int int int int int int} void \
    [ffidl-symbol [ffidl-find-lib tk] Tk_PhotoPutZoomedBlock]
ffidl-proc ffidl-photo-get-image {Tk_PhotoHandle pointer-var} int \
    [ffidl-symbol [ffidl-find-lib tk] Tk_PhotoGetImage]
ffidl-proc ffidl-photo-blank {Tk_PhotoHandle} void \
    [ffidl-symbol [ffidl-find-lib tk] Tk_PhotoBlank]
ffidl-proc ffidl-photo-expand {Tk_PhotoHandle int int} void \
    [ffidl-symbol [ffidl-find-lib tk] Tk_PhotoExpand]
ffidl-proc ffidl-photo-get-size {Tk_PhotoHandle pointer-var pointer-var} void \
    [ffidl-symbol [ffidl-find-lib tk] Tk_PhotoGetSize]
ffidl-proc ffidl-photo-set-size {Tk_PhotoHandle int int} void \
    [ffidl-symbol [ffidl-find-lib tk] Tk_PhotoSetSize]

# use the ffidl-info format for Tk_PhotoImageBlock to get the fields
proc ffidl-photo-block-fields {pib} {
    binary scan $pib [ffidl-info format Tk_PhotoImageBlock] \
	pixelPtr width height pitch pixelSize red green blue reserved
    list $pixelPtr $width $height $pitch $pixelSize $red $green $blue $reserved
}
# define accessors for the fields
foreach {name offset} {
    pixelPtr 0 width 1 height 2 pitch 3 pixelSize 4 red 5 green 6 blue 7 reserved 8
} {
    proc ffidl-photo-block.$name {pib} "lindex \[ffidl-photo-block-fields \$pib] $offset"
}

proc ffidl-photo-get-block-bytes {block} {
    set nbytes [expr {[ffidl-photo-block.height $block]*[ffidl-photo-block.pitch $block]}]
    set bytes [binary format x$nbytes]
    ffidl-memcpy bytes [ffidl-photo-block.pixelPtr $block] $nbytes
    set bytes
}

		